/*--------------------------------------------------------------------------
     "bsmiles.c" - bind SMILES notations into thermodata
        This is a part of KUCRS:
           Knowledge-basing Utilities for Complex Reaction Systems
                        (c) 2017-2021 by A. Miyoshi, Hiroshima University
                                       created: Sep.  21, 2017
                                   last edited: Aug.  19, 2021
----------------------------------------------------------------------------
  usage: bsmiles thermo.dat smiles.csv output.dat
--------------------------------------------------------------------------*/

#ifndef _CRT_SECURE_NO_WARNINGS
#define _CRT_SECURE_NO_WARNINGS
#endif

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#define MXLenFN 256             /* maximum length of file names */
#define MXLenSN 19              /* maximum length of species name */
#define MXLenSM 512             /* maximum length of smiles notation */
#define MXLen1L 1024            /* maximum length of one line */
#define NORM    0               /* normal end */
#define ERRE    (-1)            /* error end */
#define TRUE    (-1)
#define FALSE   0

int prs1sml(char *spn, char *sml, char *str);
int spn1thm(char *spn, char *str);
int getlncn(FILE *inf, char *inb);
int getlrmc(FILE *inf, char *inb);
int getline(FILE *inf, char *inb);

char buf[MXLen1L];

/* ===== main ======================================================= */

int main(int argc, char *argv[]) {
 char inFile[MXLenFN], smFile[MXLenFN], ouFile[MXLenFN];
 char thms[4][MXLen1L], smls[MXLen1L];
 char spcnth[MXLenSN], spcnsm[MXLenSN], smiles[MXLenSM];
 int ith, eof = FALSE;
 FILE *inf, *smf, *ouf;
  if (argc != 4) {
    printf(" Usage: bsmiles thermo.dat smiles.csv output.dat\n");
    return 1;
  }
  strcpy(inFile, argv[1]); strcpy(smFile, argv[2]);
  strcpy(ouFile, argv[3]);
  if ((inf = fopen(inFile, "rt")) == NULL)
    { printf(" Failed to open %s!\n", inFile); return 1; }
  if ((smf = fopen(smFile, "rt")) == NULL)
    { printf(" Failed to open %s!\n", smFile); return 1; }
  if ((ouf = fopen(ouFile, "wt")) == NULL) {
    printf(" Failed to open %s!\n", ouFile);
    fclose(inf); fclose(smf); return 1;
  }
  fprintf(ouf, "thermo all\n   300.000  1000.000  5000.000\n");
  do {
    if (getlncn(inf, thms[0]) == ERRE) { eof = TRUE; break; }
    if (spn1thm(spcnth, thms[0]) == ERRE) return 1;
    for (ith = 1; ith < 4; ith++)
      { if (getlncn(inf, thms[ith]) == ERRE) return 1; }
    if (getline(smf, smls) == ERRE) return 1;
    if (prs1sml(spcnsm, smiles, smls) == ERRE) return 1;
    if (strcmp(spcnth, spcnsm) != 0) {
      printf("Species name mismatch %s(th) %s(sml)\n", spcnth, spcnsm );
      return 1;
    }
    fprintf(ouf, "!\n");
    fprintf(ouf, "! [_ SMILES=\"%s\" _]\n", smiles);
    for (ith = 0; ith < 4; ith++) fprintf(ouf, "%s\n", thms[ith]);
  } while (!eof);
  if (!eof) return 1;
  fprintf(ouf, "end\n");
  fclose(inf); fclose(smf); fclose(ouf);
  return 0;
}

/* ----- parse one smiles input line -------------------------------- */

int prs1sml(char *spn, char *sml, char *str) {
char c;
int numq = 0;
size_t lens, pos, pq[4], lspn = 0, lsml = 0;
  lens = strlen(str);
  for (pos = 0; pos < lens; pos++) {
    if (str[pos] == '\"') { pq[numq] = pos; numq++; }
  }
  if (numq < 4) return ERRE;
  for (pos = pq[0] + 1; pos < pq[1]; pos++) {
    c = str[pos];
    if (c != ' ') { spn[lspn] = c; lspn++; }
  }
  spn[lspn] = '\0';
  if (lspn <= 0) return ERRE;
  for (pos = pq[2] + 1; pos < pq[3]; pos++) {
    c = str[pos];
    if (c != ' ') { sml[lsml] = c; lsml++; }
  }
  sml[lsml] = '\0';
  if (lsml <= 0) return ERRE;
  return NORM;
}

/* ----- extract species name from the first line of thermo data ---- */

int spn1thm(char *spn, char *str) {
char c;
size_t lens, lspn, pos;
  lens = strlen(str);
  lspn = 18;
  if (lens < lspn) lspn = lens;
  spn[lspn] = '\0';
  for (pos = 0; pos < lspn; pos++) {
    c = str[pos];
    if (c == ' ') { spn[pos] = '\0'; break; }
    else { spn[pos] = c; }
  }
  lspn = strlen(spn);
  if (lspn <= 0) return ERRE;
  return NORM;
}

/* ----- get one non-comment non-null line -------------------------- */

int getlncn(FILE *inf, char *inb) {
size_t lens;
  do {
    if (getlrmc(inf, inb) == ERRE) return ERRE;
    lens = strlen(inb);
  } while (lens <= 0);
  return NORM;
}

/* ----- get one line removing comment (after "!") ------------------ */

int getlrmc(FILE *inf, char *inb) {
size_t lens, pos;
  if (getline(inf, inb) == ERRE) return ERRE;
  lens = strlen(inb);
  for (pos = 0; pos < lens; pos++) {
    if (inb[pos] == '!') { inb[pos] = '\0'; break; }
  }
  lens = strlen(inb);
  for (pos = lens - 1; pos > 0; pos--) {
    if (inb[pos] == ' ') inb[pos] = '\0';
    else break;
  }
  return NORM;
}

/* ----- get one line ----------------------------------------------- */

int getline(FILE *inf, char *inb) {
size_t lens;
  if (fgets(inb, MXLen1L, inf) == NULL) return ERRE;
  lens = strlen(inb);
  if (lens > 0) {
    if (inb[lens - 1] == '\n') inb[lens - 1] = '\0';
  }
  return NORM;
}

