/*----------------------------------------------------------------------------
     "element.cc" - element information classes
        This is a part of KUCRS:
           Knowledge-basing Utilities for Complex Reaction Systems
                        (c) 2002-2010 by A. Miyoshi, University of Tokyo
                  separated from "molecule.cc": Oct.  30, 2003
                                        edited: Dec.  10, 2003
                                   last edited: Feb.  21, 2010
----------------------------------------------------------------------------*/

#include "element.h"

// ===== 'elementTable' and related classes ================================

// ----- element ---------------------------------------------------------

element::element() { clear(); }                        // --- constructors

element::element(const element &el) { *this = el; }

void element::clear()                                         // --- clear
  { atomicSymbol = ""; valence = 0; hyperValence = 0; formOrder = 10000; }

int element::restoreFrom(ifstream &ifs) {               // --- restoreFrom
  vector<string> vs;
  if (getTokensFrom(ifs, 6, vs) == ERRE) { return ERRE; }
   if (vs[0] != "element{") { return ERRE; }
   atomicSymbol = vs[1];
   valence = atoi(vs[2].c_str());
   hyperValence = atoi(vs[3].c_str());
   formOrder = atoi(vs[4].c_str());
   if (vs[5] != "}") { return ERRE; }
  return NORM;
}

string element::canonSymbol() {                         // --- canonSymbol
  // returns an atomic symbol in canonical form, i.e. 1st letter is
  //  in uppercase and second lowercase.
  string cansym = atomicSymbol;
  if (cansym.length() > 1) { cansym[1] = tolower(cansym[1]); }
  return cansym;
}

// ----- elementTable ----------------------------------------------------

// object is loaded from FL_ELMTBL in the default constructor

elementTable::elementTable() { init(); }                // --- constructor

void elementTable::init() {                                    // --- init
  libPath libp;
  clear(); none.clear();
  aufbTbl.clear();
  aufbTbl[0] = "1s";
  aufbTbl[2] = "2s"; aufbTbl[4] = "2p";
  aufbTbl[10] = "3s"; aufbTbl[12] = "3p";
  aufbTbl[18] = "4s"; aufbTbl[20] = "3d"; aufbTbl[30] = "4p";
  aufbTbl[36] = "5s"; aufbTbl[38] = "4d"; aufbTbl[48] = "5p";
  aufbTbl[54] = "6s"; aufbTbl[56] = "4f"; aufbTbl[70] = "5d"; aufbTbl[80] = "6p";
  aufbTbl[86] = "7s"; aufbTbl[88] = "5f";
  aufbUpLimit = 102;
  showRevision();
  if (loadFrom(FL_ELMTBL, libp) == ERRE)
    { cout << "Error reading elements.\n"; }
}

int elementTable::loadFrom(string flname, libPath &libp) { // --- loadFrom
  int atmcNbr, endTable = false;
  string s;
  element el;
  ifstream ifs(libp.fullPath(flname).c_str());
  if (!ifs) { return ERRE; }
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "elementTable{") { return ERRE; }
  if (!empty()) { clear(); }
  do {
    if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
    if (s == "}") { endTable = true; }
    else {
      atmcNbr = atoi(s.c_str());
      if (el.restoreFrom(ifs) == ERRE) { return ERRE; }
      insert(make_pair(atmcNbr, el));
    }
  } while (!endTable);
  ifs.close();
  cout << size() << " elements loaded from [" << flname << "]\n";
  return NORM;
}

int elementTable::atmcNbrBySymb(const string &as) {   // --- atmcNbrBySymb
  iterator p;
  string asym = strToUpper(as);
  for (p = begin(); p != end(); p++) {
    if (p->second.atomicSymbol == asym) { return p->first; }
  }
  return (-1);
}

int elementTable::isSupported(int an) {                 // --- isSupported
  if (find(an) == end()) { return false; }
  return true;
}

const element &elementTable::operator[](int an) {        // --- operator[]
  iterator p;
  if ((p = find(an)) == end()) {
    cout << "\nERROR elementTable [unsupported element #"
     << an << "]\n\n";
    return none;
  }
  return p->second;
}
                                                         // --- ionValence
int elementTable::ionValence(int an, int chg, int &hyp) {
  iterator p;
  string orb;
  int eqAn = an - chg, nvel;
  hyp = 0;
  if (!isSupported(an)) {
    cout << "\nERROR elementTable [unsupported element #"
     << an << "]\n\n";
    return 0;
  }
  if ((p = find(eqAn)) != end()) {
    hyp = p->second.hyperValence;
    return p->second.valence;
  }
  if ((nvel = aufbauValence(eqAn, orb)) < 0) {
    cout << "\nERROR elementTable [invalid charge: " << chg << "]\n\n";
    return 0;
  }
  if (orb == "1s") { return min(nvel, 2 - nvel); }
  if (orb[1] == 's') { return nvel; }
  if (orb[1] == 'p') { return min(nvel + 2, 6 - nvel); }
  cout << "\nERROR elementTable [unsupported electron configulation"
   << " for " << an << "electons\n\n";
  return 0;
}

string elementTable::canonSymbol(int an) {              // --- canonSymbol
  // canonical atomic symbol (1st letter uppercase / second lowercase)
  iterator p;
  if ((p = find(an)) == end()) { return ""; }
  return p->second.canonSymbol();
}

                                                      // --- aufbauValence
int elementTable::aufbauValence(int an, string &orbit) {
  int nvel = 0;
  map<int, string>::reverse_iterator p;
  orbit = "1s";
  if ((an < 0) || (an > aufbUpLimit)) { return -1; }
  for (p = aufbTbl.rbegin(); p != aufbTbl.rend(); p++) {
    if (p->first < an)
      { nvel = an - p->first; orbit = p->second; break; }
  }
  return nvel;
}

// ----- globals ---------------------------------------------------------

string rpsAtmAttr(int atmAttr) {
  static string rpsa[5] = {" ", ".", ":", "!", "#"};
  string s = "?";
  if ((atmAttr < 0) || (atmAttr >= 5))
    { cout << "ERROR rpsAtmAttr [undefined]\n"; return s; }
  s = rpsa[atmAttr];
  return s;
}

string rpsBndOrdr(int bndOrdr) {
  static string rpsb[4] = {" ", "-", "=", "%"};
  string s = "?";
  if ((bndOrdr < 0) || (bndOrdr >= 4))
    { cout << "ERROR rpsBndOrdr [undefined]\n"; return s; }
  s = rpsb[bndOrdr];
  return s;
}

// ===== 'atmcComp' and related classes ==================================

// ----- atmcComp --------------------------------------------------------

// ---------------------------------------------------------- [operator+=]
atmcComp &atmcComp::operator+=(atmcComp &ac) {
  *this = *this + ac; return *this;
}

// ---------------------------------------------------------- [operator==]
int atmcComp::operator==(atmcComp &ac) {
  iterator p, q;
  if (size() != ac.size()) { return false; }
  for (p = begin(), q = ac.begin(); p != end(); p++, q++)
    { if (*p != *q) { return false; } }
  return true;
}

// ---------------------------------------------------------- [operator!=]
int atmcComp::operator!=(atmcComp &ac) { return !(*this == ac); }

// ----------------------------------- [operator+(atmcComp &, atmcComp &)]
atmcComp operator+(atmcComp &acl, atmcComp &acr) {
  atmcComp acsum = acl;
  for (atmcComp::iterator p = acr.begin(); p != acr.end(); p++)
    { acsum[p->first] += p->second; }
  return acsum;
}

// ----- atmcCompSorted --------------------------------------------------

// ------------------------------------------------------------- [setFrom]
void atmcCompSorted::setFrom(atmcComp &ac) {
  atmcComp::iterator p;
  clear();
  for (p = ac.begin(); p != ac.end(); p++)
    { insert(make_pair(elmt[p->first].formOrder, *p)); }
}

// ===== global object of elementTable ===================================

elementTable elmt;

