/*--------------------------------------------------------------------------
     "moleclst.cc" - molecule list administration classes
        This is a part of KUCRS:
           Knowledge-basing Utilities for Complex Reaction Systems
                        (c) 2002-2010 by A. Miyoshi, University of Tokyo
                                       created: Sept. 12, 2002
                                        edited: Dec.  10, 2003
                                   last edited: Feb.  20, 2010
--------------------------------------------------------------------------*/

#include "moleclst.h"

/*****************************************************************************
     'nameControl' and related classes
*****************************************************************************/


/*============================================================================
     'namePool' class
============================================================================*/


// ------------------------------------------------------------ [register]
int namePool::regName(string name) {
  if (isUsed(name)) { return ERRE; }
  insert(name);
  return NORM;
}

// --------------------------------------------------- [getNamesBeginWith]
int namePool::getNamesBeginWith(string begn, vector<string> &names) {
  int lenbn = begn.length(), numFound = 0;
  if (!names.empty()) { names.clear(); }
  for (iterator p = lower_bound(begn); p != end(); p++) {
    if (p->substr(0, lenbn) == begn) { names.push_back(*p); numFound++; }
    else { break; }
  }
  return numFound;
}

// --------------------------------------------------------------- [print]
void namePool::print(ostream &os) {
  os << "----- namePool -----\n";
  for (iterator p = begin(); p != end(); p++)
    { os << "[" << *p << "]\n"; }
}

// -------------------------------------------------------------- [isUsed]
int namePool::isUsed(string name) {
  if (find(name) != end()) { return true; } else { return false; }
}

// --------------------------------- [operator << (ostream &, namePool &)]
ostream &operator<<(ostream &os, namePool &nl)
  { nl.print(os); return os; }


/*============================================================================
     'baseNameCntrl' class
============================================================================*/


// -------------------------------------------------------- [updateByName]
//   update basename library from a name

void baseNameCntrl::updateByName(string name) {
  string bnam;
  iterator p;
  pair<iterator, iterator> pp = equal_range(name);
  if (pp.first != pp.second) {     // - basename == name found
    if ((pp.first)->second < 1) { (pp.first)->second = 1; }
  } else if (pp.first != begin()) {
    p = pp.first; p--; bnam = p->first;
    if (name.substr(0, bnam.length()) == bnam) { (p->second)++; }
  }
}


// ------------------------------------------------------- [updateGetAuto]
//   update and get new suffix for basename

string baseNameCntrl::updateGetAuto(string bnam, namePool &nl) {
  iterator p = find(bnam);
  if (p == end()) { p = setNewBasename(bnam, nl); }
  string sfx = idxToSfx(p->second); (p->second)++;
  return sfx;
}

// ----------------------------------------------------- [updateByBnamSfx]
//   update by basename and suffix

void baseNameCntrl::updateByBnamSfx(string bnam, string sfx,
 namePool &nl) {
  iterator p = find(bnam);
  if (p == end()) { p = setNewBasename(bnam, nl); }
  int idx = sfxToIdx(sfx);
  if (idx >= p->second) { p->second = (idx + 1); }
}

// --------------------------------------------------------------- [print]
void baseNameCntrl::print(ostream &os) {
  os << "----- baseNameCntrl -----\n";
  for (iterator p = begin(); p != end(); p++) {
    os << "[" << p->first << "] " << p->second << endl;
  }
}

// ------------------------------------------------------ [setNewBasename]
baseNameCntrl::iterator baseNameCntrl::setNewBasename(string bnam,
 namePool &nl) {
  vector<string> namesFound;
  int numFound = nl.getNamesBeginWith(bnam, namesFound);
  int lenBn = bnam.length(), idx, maxIdx = 0;
  for (int il = 0; il < numFound; il++) {
    idx = sfxToIdx(namesFound[il].substr(lenBn));
    if (idx > maxIdx) { maxIdx = idx; }
  }
  return insert(make_pair(bnam, (maxIdx + 1))).first;
}

// ------------------------------------------------------------ [idxToSfx]
//   Get suffix corresponding to the index

string baseNameCntrl::idxToSfx(int idx) {
  static int baseNum = 26;
  static char digit[27] = "abcdefghijklmnopqrstuvwxyz";
  int numDigs = -1, maxByDig = 0, minByDig, powOfbase = 1, idxrem, iDig;
  int rem, quot;
  string rets = "";

  if (idx < 0)
    { cout << "ERROR baseNameCntrl::idxToSfx [idx < 0]\n"; return rets; }
  do {
    numDigs++; minByDig = maxByDig;
    maxByDig += powOfbase; powOfbase *= baseNum;
  } while (idx >= maxByDig);
  idxrem = idx - minByDig;
  for (iDig = 0; iDig < numDigs; iDig++) {
    quot = idxrem / baseNum; rem = idxrem % baseNum;
    rets = digit[rem] + rets;
    idxrem = quot;
  }
  return rets;
}

// ------------------------------------------------------------ [sfxToIdx]
//   Get index of the suffix

int baseNameCntrl::sfxToIdx(string sfx) {
  static const int baseNum = 26;
  static const char digit[27] = "abcdefghijklmnopqrstuvwxyz";
  int numDigs = sfx.length(), iDig, powOfbase = 1, minByDig = 0;
  int idxrem = 0, ic, digval;
  char c;

  for (iDig = 0; iDig < numDigs; iDig++)
    { minByDig += powOfbase; powOfbase *= baseNum; }
  for (iDig = 0; iDig < numDigs; iDig++) {
    c = sfx[iDig]; digval = -1;
    for (ic = 0; ic < baseNum; ic++)
      { if (digit[ic] == c) { digval = ic; break; } }
    if (digval < 0) { return -1; }
    idxrem *= baseNum; idxrem += digval;
  }
  return (minByDig + idxrem);
}

// ---------------------------- [operator << (ostream &, baseNameCntrl &)]
ostream &operator<<(ostream &os, baseNameCntrl &bl)
  { bl.print(os); return os; }


/*============================================================================
     'nameControl' class
============================================================================*/


// --------------------------------------------------------------- [clear]
void nameControl::clear()
  { nameLib.clear(); baseLib.clear(); }

// ------------------------------------------------------------- [regName]
//   Register a name. Returns the input on success. Returns a new name
//    starts with bnam on failure.

string nameControl::regName(string name, string bnam) {
  if (nameLib.regName(name) == NORM)
    { baseLib.updateByName(name); return name; }
  return regBaseAuto(bnam);
}

// ---------------------------------------------------------- [regBaseSfx]
//   Register a name = bnam+sfx. Returns the input name on success.
//    Returns a name with another suffix on failure.

string nameControl::regBaseSfx(string bnam, string sfx) {
  string name = bnam + sfx;
  if (nameLib.regName(name) == NORM) {
    baseLib.updateByBnamSfx(bnam, sfx, nameLib);
    return name;
  }
  return regBaseAuto(bnam);
}

// --------------------------------------------------------- [regBaseAuto]
//   Get and register a new name beginning with bnam.

string nameControl::regBaseAuto(string bnam) {
  string sfx = baseLib.updateGetAuto(bnam, nameLib);
  string name = bnam + sfx;
  if (nameLib.regName(name) == ERRE)
    { cout << "ERROR nameControl::regBaseAuto [unexpected]\n"; }
  return name;
}

// --------------------------------------------------------------- [print]
void nameControl::print(ostream &os)
  { nameLib.print(os); baseLib.print(os); }

// ------------------------------ [operator << (ostream &, nameControl &)]
ostream &operator<<(ostream &os, nameControl &nc)
  { nc.print(os); return os; }


/*****************************************************************************
     'molecList' and related classes
*****************************************************************************/


/*============================================================================
     'molecRef' class
============================================================================*/


// ---------------------------------------------------------- [molecRef()]
molecRef::molecRef() { clear(); }

// ------------------------------------ [molecRef(molecule &, groupLib &)]
molecRef::molecRef(molecule &m, groupLib &gl) { setFrom(m, gl); }

// --------------------------------------------------------------- [clear]
void molecRef::clear()
  { name = ""; molec.clear(); isRxnDone = false; chemForm = ""; }

// --------------------------------------------------------- [restoreFrom]
int molecRef::restoreFrom(ifstream &ifs, nameControl &nc, groupLib &gl) {
  molecule mol;
  if (mol.restoreFrom(ifs) == ERRE) { return ERRE; }
  setFrom(mol, gl);
  string rnam = nc.regName(name, chemForm);
  if (rnam != name) {
    cout << "WARNING name [" << name << "] already used. changed to ["
         << rnam << "]\n";
    name = rnam;
  }
  return NORM;
}

// --------------------------------------------------------- [isRO2ofDone]
int molecRef::isRO2ofDone() {
  if (!molec.isPeroxyRad()) { return false; }
  if ((name == "HO2") || (name == "CH3OO")) { return false; }
  if (!isRxnDone) { return false; }
  return true;
}

// ------------------------------------------------------------- [setFrom]
void molecRef::setFrom(molecule &m, groupLib &gl) {
  isRxnDone = m.isRxnDone;
  if (isRxnDone) {
    if (!m.isUnique) { m.beUnique(); }
    mtyp = m.getType(gl);
  } else { mtyp = 0L; }
  name = m.molName; molec = m.molec; chemForm = molec.getChemFormula();
}


/*============================================================================
     'molecList' class
============================================================================*/


// ---------------------------------------------------------- [appendFrom]
//   append molecRef from a file  (# does not clear itself)

int molecList::appendFrom(string flname, libPath &libp, nameControl &nc,
 groupLib &gl) {
  int nRef = size();
  molecRef mr;
  ifstream ifs(libp.fullPath(flname).c_str());
  if (!ifs) { return ERRE; }
  while (mr.restoreFrom(ifs, nc, gl) == NORM) { push_back(mr); }
  ifs.close();
  cout << (size() - nRef) << " molecules read from [" << flname << "]\n";
  return NORM;
}

// ---------------------------------------------- [findIdxOf(molecInfo &)]
int molecList::findIdxOf(molecInfo &m) {
  int nRefL, ir;
  string mChmFrm;
  mChmFrm = m.getChemFormula(); nRefL = size();
  for (ir = 0; ir < nRefL; ir++) {
    if ((*this)[ir].chemForm == mChmFrm) {
      if ((*this)[ir].molec == m) { return ir; }
    }
  }
  return -1;
}

// ----------------------------------------------- [findIdxOf(molecule &)]
int molecList::findIdxOf(molecule &m)
  { molecInfo mi = m.molec; return findIdxOf(mi); }

// --------------------------------------------------- [find(molecInfo &)]
molecList::iterator molecList::find(molecInfo &m) {
  iterator p;
  string mChmFrm;
  mChmFrm = m.getChemFormula();
  for (p = begin(); p != end(); p++) {
    if (p->chemForm == mChmFrm) {
      if (p->molec == m) { return p; }
    }
  }
  return end();
}

// ---------------------------------------------------- [find(molecule &)]
molecList::iterator molecList::find(molecule &m)
  { molecInfo mi = m.molec; return find(mi); }

// -------------------------------------------------------- [find(string)]
molecList::iterator molecList::find(string nam) {
  for (iterator p = begin(); p != end(); p++) {
    if (p->name == nam) { return p; }
  }
  return end();
}

// ---------------------------------------------- [removeMol(molecInfo &)]
void molecList::removeMol(molecInfo &m) {
  iterator p = find(m);
  if (p == end()) { return; }
  if (p->isRxnDone) { return; }
  erase(p);
}

// ----------------------------------------------- [removeMol(molecule &)]
void molecList::removeMol(molecule &m)
  { molecInfo mi = m.molec; removeMol(mi); }


/*============================================================================
     'molecInQue' class
============================================================================*/


// -------------------------------------------------------- [molecInQue()]
molecInQue::molecInQue() { clear(); }

// ---------------------------------------------- [molecInQue(molecule &)]
molecInQue::molecInQue(molecule &m) { setFrom(m); }

// --------------------------------------------------------------- [clear]
void molecInQue::clear() { mol.clear(); chemForm = ""; }

// ------------------------------------------------------------- [setFrom]
void molecInQue::setFrom(molecule &m)
  { mol = m; chemForm = mol.molec.getChemFormula(); }

// --------------------------------------------------------------- [print]
void molecInQue::print(ostream &os)
  { os << mol.molName << " [" << chemForm << "]\n"; }

// ------------------------------- [operator << (ostream &, molecInQue &)]
ostream &operator<<(ostream &os, molecInQue &miq)
  { miq.print(os); return os; }


/*============================================================================
     'molecQueue' class
============================================================================*/

// ---------------------------------------------------------- [appendFrom]
void molecQueue::appendFrom(molecQueue &mq)
  { for (iterator p = mq.begin(); p != mq.end(); p++) { push_back(*p); } }

// ---------------------------------------------- [findIdxOf(molecInfo &)]
int molecQueue::findIdxOf(molecInfo &m) {
  int nRefL, ir;
  string mChmFrm;
  mChmFrm = m.getChemFormula(); nRefL = size();
  for (ir = 0; ir < nRefL; ir++) {
    if ((*this)[ir].chemForm == mChmFrm) {
      if ((*this)[ir].mol.molec == m) { return ir; }
    }
  }
  return -1;
}

// ----------------------------------------------- [findIdxOf(molecule &)]
int molecQueue::findIdxOf(molecule &m) {
  molecInfo mi = m.molec;
  return findIdxOf(mi);
}

// --------------------------------------------------- [find(molecInfo &)]
molecQueue::iterator molecQueue::find(molecInfo &m) {
  iterator p;
  string mChmFrm;
  mChmFrm = m.getChemFormula();
  for (p = begin(); p != end(); p++) {
    if (p->chemForm == mChmFrm) {
      if (p->mol.molec == m) { return p; }
    }
  }
  return end();
}

// ---------------------------------------------------- [find(molecule &)]
molecQueue::iterator molecQueue::find(molecule &m) {
  molecInfo mi = m.molec;
  return find(mi);
}

// -------------------------------------------------------- [getPopQueTop]
int molecQueue::getPopQueTop(molecule &m) {
  iterator p = begin();
  if (p == end()) { return ERRE; }
  m = p->mol; pop_front();
  return NORM;
}

// --------------------------------------------------------------- [print]
void molecQueue::print(ostream &os) {
  os << "----- molecQueue -----\n";
  for (iterator p = begin(); p != end(); p++) { os << *p; }
}

// ------------------------------- [operator << (ostream &, molecInQue &)]
ostream &operator<<(ostream &os, molecQueue &mq)
  { mq.print(os); return os; }


/*============================================================================
     'specAdmin' class
============================================================================*/

// --------------------------------------------------------------- [clear]
void specAdmin::clear() { namAd.clear(); molAd.clear(); rxnQue.clear(); }

// --------------------------------------------- [getMolStat(molecInfo &)]
int specAdmin::getMolStat(molecInfo &m, string &nam) {
  /* -------------------------------------
   return code:
    0 : not found anywhere
    1 : found in reaction queue
    2 : name found in library
    3 : found in library (reaction done)
  ------------------------------------- */
  molecList::iterator p = molAd.find(m);
  if (p != molAd.end()) {
    nam = p->name;
    if (p->isRxnDone) { return 3; }
    return 2;
  }
  molecQueue::iterator q = rxnQue.find(m);
  if (q != rxnQue.end()) { nam = q->mol.molName; return 1; }
  return 0;
}

// ---------------------------------------------- [getMolStat(molecule &)]
int specAdmin::getMolStat(molecule &m, string &nam)
  { molecInfo mi = m.molec; return getMolStat(mi, nam); }

