/*--------------------------------------------------------------------------
     "rxnmacro.h" - reaction rate macros
        This is a part of KUCRS:
           Knowledge-basing Utilities for Complex Reaction Systems
                        (c) 2002-2010 by A. Miyoshi, University of Tokyo
                                       created: Oct.  31, 2003
                                        edited: Dec.  10, 2003
                                        edited: July   7, 2010
                                   last edited: Nov.  19, 2010
                            - added 'sum-of-rates' expression
--------------------------------------------------------------------------*/

#ifndef _RXNMACRO_H_
#define _RXNMACRO_H_

#include "element.h"

#define FL_RXNMCR "rxn_macr.dat"        // reaction rate macros

// --- constants used in the previous version of genReact
// #define RGASCAL   1.987216
// #define RATMMOL   82.0578

#define RGASCAL   1.9872065        // gas constant in cal K-1 mol-1
#define RATMMOL   82.057459        // gas constant in atm K-1 cm3 mole-1
#define NUMTLST   13               // number of temperatures in list

// ***** rateMacro and related classes ***********************************

class rxnRate {
 public:                                         // --- data members ---
  double A, b, Ea;
  int isTroe;
  double AL, bL, EaL, aFc, T3Fc, T1Fc, T2Fc;
 public:                                     // --- member functions ---
  rxnRate();
  rxnRate(const rxnRate &rate);
  void clear();
  int restoreFrom(ifstream &ifs);
  int restoreLowTroeFrom(ifstream &ifs);
  string format();
  string formatLow();
  string formatTroe();
};

class rateMacro : public map<string, rxnRate> {
 public:                                     // --- member functions ---
  rateMacro();
  int loadFrom(string flname, libPath &libp);
};

// ***** thermNASA and related classes ***********************************

class thermNASA {
 public:                                         // --- data members ---
  int isUsed;
  atmcComp atmCmp;
  double lowT, highT, comT;
  double c[2][7];
 public:                                     // --- member functions ---
  thermNASA();
  thermNASA(const thermNASA &tm);
  void clear();
  int restoreFrom(vector<string> &vs);
  void getThermFuncs(double T, double &HRT, double &SR, double &CpR);
  double getValCoeff(const string &s);
  string fixCoefField(const string &s);
  string fix1stLine(const string &s, int asLower);
  string fixDataLine(const string &s, int lineNbr);
};

class thermTable : public map<string, thermNASA> {
 public:                                     // --- member functions ---
  void setM();
  int appendFrom(string flname);
  int appendFrom(string flname, libPath &libp);
  void setAllIsUsed();
  set<int> getUsedAtoms();
 private:
  int append(string flname);
};

// ***** rxnInfo and related classes *************************************

class rxnInfo {
 public:                                         // --- data members ---
  vector<string> reac;  // reactant names
  vector<string> prod;  // product names
  string rate;          // rate macro expression
                           // --- sum-of-rates (Nov. 19, 2010)
  int isSumRates;          // flag to indicate sum-of-rates
  vector<double> factRate; // multiplication factor
  vector<string> rateStr;  // rate macro expressions
                           // ---
  double shiftE;        // shift of Ea
  double factA;         // multiplication factor for A
  double shiftB;        // shift of b
  int isRevrs;          // flag to indicate reversible reaction
  int givenRev;         // flag to indicate that rate is given for rev.
  int modTroe;          // flag to indicate Troe mod pars. exist
  double modAL, modbL, modEaL, modaFc, modT3Fc, modT1Fc, modT2Fc;
  double delH, delS;    // delta-H [kcal mol-1], delta-S [cal K-1 mol-1]
  rxnRate kf;           // forward rate const.
  rxnRate kr;           // reverse rate const.
  double kfTbl[NUMTLST];
  double krTbl[NUMTLST];
  double KeTbl[NUMTLST];
 public:                                     // --- member functions ---
  rxnInfo();
  void clear();
  int restoreFrom(vector<string> &vs, thermTable &t, rateMacro &m);
  int setRateConst(thermTable &t, rateMacro &m, double Ts[]);
  void outputToChem(ofstream &ofs, double Ts[]);
 private:
  int parseSumRates(string &s, rateMacro &m);
  int setSumOfRates(thermTable &t, rateMacro &m, double Ts[]);
  void setTherm298(thermTable &t);
  int calRatesEquil(thermTable &t, double Ts[]);
  void calRatesFowerdOnly(double Ts[]);
  int fitRevRate(double Ts[]);
  void fitRRsetMat(double q[], double B[][6], double Ts[]);
  int fitRRinvB(double B[][6]);
  void fitRRcalA(double q[], double B[][6], double a[]);
};

#endif
